@@ -12,6 +12,7 @@ from .test import test
 
 
 def _temperature_fn(max_training_steps, training_steps):
+  r"""Determines the randomness for the action taken by the model"""
   if training_steps < 0.5 * max_training_steps:
       return 1.0
   elif training_steps < 0.75 * max_training_steps:
@@ -27,7 +28,7 @@ def fit(model, env_id,
           test_interval: int = 10,
           num_test_episodes: int = 10,
           max_training_steps: int = 10000,
-          save_every_n_steps: int = 1000,
+          save_every_n_epochs: int = 1,
           num_simulations: int = 50,
           k_steps: int = 10,
           buffer_warm_up: int = 128,
@@ -41,6 +42,60 @@ def fit(model, env_id,
           temperature_fn=_temperature_fn,
           log_all_metrics=False,
           ):
+  r"""  Fits the model on the given `env_id` environment.
+        
+        Parameters
+        ----------
+        model: An instance of `MuZero`
+        
+        env_id: str, the environment id for `gym.make(env_id, render_mode='rgb_array')`
+        
+        tracer: An instance of episode tracer that inherents `BaseTracer`. 
+        
+        buffer: An instance of replay buffer that inherents `BaseReplayBuffer`
+
+        max_episodes: int, positive integer. Maximum training episodes
+
+        test_interval: int, positive integer. Tests the model every `test_interval` episodes.
+
+        num_test_episodes: int, positive integer. The number of episodes to test the model.
+
+        max_training_steps: int, positive integer. Maximum training steps.
+
+        save_every_n_epochs: int, positive integer. Save the model every `save_every_n_epochs` episodes.
+
+        num_simulations: int, positive int. The number of simulations.
+
+        k_steps: int, positive int. The number of unrolled steps.
+
+        buffer_warm_up: int, positive int. Collected trajectories until at least `buffer_warm_up` 
+        trajectories are stored in the buffer.
+
+        num_trajectory: int, positive int. Number of trajectories to sample from the buffer.
+
+        sample_per_trajectory: int, positive int. The number of training samples from each of the trajectories.
+
+        name: str. The name of the `TrainMonitor`, and the tensorboard is stored at 
+        `tensorboard_dir=os.path.join(tensorboard_dir, name)`
+
+        tensorboard_dir: str. The directory to store the tensorboard.
+
+        model_save_path: str. The path to save the model parameters during training.
+
+        save_name: str. The name of the file to store the model parameters.
+
+        random_seed: int. random seed
+
+        temperature_fn: Callable. Determines the randomness for the action taken by the model
+
+        log_all_metrics: bool. If True, all metrics will be displayed by the `TrainMonitor`; 
+        else only `T`, `avg_r`, `G`, `avg_G`, `t`, `dt` will be displayed.
+        
+        Returns
+        -------
+        model_path: str. A path to the model parameter that has the best performance during testing.
+
+  """
   if name is None:
     name = env_id 
   if tensorboard_dir is None:
@@ -73,6 +128,7 @@ def fit(model, env_id,
   print('buffer warm up stage...')
   while len(buffer) < buffer_warm_up:
     obs, info = env.reset()    
+    tracer.reset()
     trajectory = Trajectory()
     temperature = temperature_fn(max_training_steps=max_training_steps, training_steps=training_step)
     for t in range(env.spec.max_episode_steps):
@@ -102,7 +158,8 @@ def fit(model, env_id,
     tensorboard_dir=os.path.join(tensorboard_dir, name),
     log_all_metrics=log_all_metrics)
   for ep in range(max_episodes):
-    obs, info = env.reset(seed=random_seed)    
+    obs, info = env.reset(seed=random_seed)   
+    tracer.reset() 
     trajectory = Trajectory()
     temperature = temperature_fn(max_training_steps=max_training_steps, training_steps=training_step)
     for t in range(env.spec.max_episode_steps):
@@ -128,24 +185,27 @@ def fit(model, env_id,
     if len(trajectory) >= k_steps:
       buffer.add(trajectory, trajectory.batched_transitions.w.mean())
 
+    train_loss = 0
     for _ in range(t):
       transition_batch = buffer.sample(num_trajectory=num_trajectory,
                                         sample_per_trajectory=sample_per_trajectory,
                                         k_steps=k_steps)
       loss_metric = model.update(transition_batch)
+      train_loss += loss_metric['loss']
       training_step += 1
-      env.record_metrics(loss_metric)
       
-      if training_step % save_every_n_steps == 0:
-        model_folder_name = f'epoch_{ep:04d}_loss_{loss_metric["loss"]:.8f}'
-        if not os.path.exists(os.path.join(model_dir, model_folder_name)):
-          os.makedirs(os.path.join(model_dir, model_folder_name))
-        cur_path = os.path.join(model_dir, model_folder_name, save_name) 
-        model.save(cur_path)
-        if not model_path:
-          model_path = cur_path
-      if training_step >= max_training_steps:
-        return model_path
+    train_loss /= t 
+    env.record_metrics({'loss': train_loss})
+    if ep % save_every_n_epochs == 0:
+      model_folder_name = f'epoch_{ep:04d}_loss_{train_loss:.8f}'
+      if not os.path.exists(os.path.join(model_dir, model_folder_name)):
+        os.makedirs(os.path.join(model_dir, model_folder_name))
+      cur_path = os.path.join(model_dir, model_folder_name, save_name) 
+      model.save(cur_path)
+      if not model_path:
+        model_path = cur_path
+    if training_step >= max_training_steps:
+      return model_path
     env.record_metrics({'training_step': training_step})
 
     # Periodically test the model
@@ -158,6 +218,7 @@ def fit(model, env_id,
         model_folder_name = f'epoch_{ep:04d}_test_G_{test_G:.8f}'
         if not os.path.exists(os.path.join(model_dir, model_folder_name)):
           os.makedirs(os.path.join(model_dir, model_folder_name))
-        model.save(os.path.join(model_dir, model_folder_name, save_name))
+        model_path = os.path.join(model_dir, model_folder_name, save_name)
+        model.save(model_path)
 
   return model_path
